import cv2
import threading
import time
import mouse
import keyboard
import pandas as pd
import pathlib
import re
import os
import numpy as np
import csv
import traceback
import torch

from collections import defaultdict, deque
from PIL import Image

from ..module import Module, MouseControl, KeyControl
from ..detector import Detector

class HAC:

    saved_images = []    # images to data
    images = deque()     # captured images
    tss = deque()        # timestamps
    controls = deque()   # controls need to be executed
    modules = {}         # available modules of this controller
    module = None        # current module
    buffer_size = 10

    def __init__(self, holistic_tracker):
        """
        Human Action Controller (HAC)
        A Human action controller needs a holistic tracker to extract a skeleton from an image.

        inputs:
            holistic_tracker: a tracker to extract a skeleton from an image
                              take hac.tracker.holistic_tracker as a reference.
        """
        self.holistic_tracker = holistic_tracker

    def add_module(self, model_name, module_name=""):
        """
        Add an module to HAC.
        
        inputs: 
            model_name: a module contains a set of actions which can be detected by a detector.
        """

        if module_name == "":
            module_name = model_name

        model_data_path = os.path.join("pyhac", "trained_model", "gcn", \
                                        model_name, "best_model.pth")
        device = "cuda" if torch.cuda.is_available() else "cpu"
        model_data = torch.load(model_data_path, map_location=device)
        detector = Detector(model_data)
        module = Module(module_name, detector)

        if module_name in self.modules:
            raise KeyError(f"module_name: {module_name} exists!")

        self.modules[module_name] = module

        return module
        
    def set_init_module(self, module):
        """
        Set a default module. Check hac.module to know more about modules
        """

        self.module = module

    def buffer_resize(self):

        if len(self.images) > self.buffer_size:
            self.images.popleft()

        if len(self.tss) > self.buffer_size:
            self.tss.popleft()

        if len(self.controls) > self.buffer_size:
            self.controls.popleft()

    def update(self, image, ts, keep_data=False):
        '''
        We call the function ```update``` every frame to capture an action from a skeleton, 
        and then generate controls.
        '''

        if self.tss and self.tss[-1] == ts:
            return

        # Alleviate jittering detection result
        image_blur = cv2.GaussianBlur(image, (7, 7), 0)

        # capture the skeleton from an image
        skeleton = self.holistic_tracker(image_blur, ts)
        
        if keep_data:
            control = None
            self.df_data = skeleton
        else:
            # skeleton -> action detection -> control
            control = self.module(skeleton)
        
        self.images.append(image)
        self.tss.append(ts)
        self.controls.append(control)

        self.buffer_resize()

    def release_keys(self):
        """
        Release keys which aren't pressed.
        """
        for _, control in self.module.mapping.items():
            if not isinstance(self.controls[-1], KeyControl):
                if isinstance(control, KeyControl):
                    control.release()
            else:
                if isinstance(control, KeyControl) and control.key != self.controls[-1].key:
                    control.release()

    def release_all(self):
        """
        Release keys and mouse to prevent from blocking devices.
        """
        for _, control in self.module.mapping.items():
            if isinstance(self.controls[-1], KeyControl) or \
               isinstance(self.controls[-1], MouseControl):
                control.release()

    def execute(self):
        """
        Execute controls (e.g. move mouse or press keys)
        """
        if len(self.controls) == 0:
            return 
        
        self.release_keys() # release not pressed keys

        if not self.controls[-1]:
            return

        if isinstance(self.controls[-1], Module): # Transition
            self.module = self.controls[-1]
            self.module.reset()
            print(f"Transit to module: {self.module.name}")
        else:
            if self.controls[-1].execute:
                self.controls[-1].execute()
        
    def save(self, csv_path, image_dir, label):
        """
        For collecting data.
        """

        csv_dir = os.path.dirname(csv_path)
        image_name = str(int(time.time() * 1000)) + ".png"
        image_path = os.path.join(image_dir, image_name)

        pathlib.Path(csv_dir).mkdir(exist_ok=True, parents=True)
        pathlib.Path(image_dir).mkdir(exist_ok=True, parents=True)

        if label is None:
            label = input("Label: ")
            label = re.findall("\d+", label)[0]

        self.df_data["image_name"] = [image_name]
        self.df_data["label"] = [label]

        image = self.images[-1]

        if os.path.exists(csv_path):
            with open(csv_path, 'a') as f:
                write = csv.writer(f)
                write.writerow(self.df_data.values[0,:].tolist())
        else:
            df_data = self.df_data
            df_data.to_csv(csv_path, index=False)
        
        self.saved_images.append((image, image_path))

    def save_images(self):
        """
        Call save_images after the end of data collection, to save all images once.
        """

        for image, path in self.saved_images:
            image_pil = Image.fromarray(image)
            image_pil.save(path)


import os 
import pathlib
import pickle
import torch
from ..utils.key_points import W_LIST_POSE, W2I_POSE, W_LIST_LEFT_HAND, W2I_LEFT_HAND, W_LIST_RIGHT_HAND, W2I_RIGHT_HAND
from ..utils.normalizer import normalize_data, normalize_hand_data
from ..model import HACModel

class Detector:

    def __init__(self, model_data):

        """
        init: 
            model: trained model 
            pred2str: a function convert predictions to string
        """
        self.model = model_data["model"]
        self.target_columns = model_data["target_columns"]
        self.mapping = model_data["actions"]
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

    def __call__(self, df):
        """
        input:
            x: inputs for model to predict hand gesture, the type of x depends on the model.
        output:
            output: hand gesture like "one", "two", "three", "stone"...

        """
        data = self.normalize(df)
        x = torch.from_numpy(data).view(-1, self.model.in_channels, self.model.get_num_vertices(), 1).float().to(self.device)
        output = self.model(x) # batch, (x, y), num of joints, t
        pred = output.reshape(-1, len(self.mapping)).cpu().argmax(axis=1, keepdim=True)
        pred_gesture = self.pred2str(pred)
        print(pred_gesture)

        return pred_gesture

    def pred2str(self, pred):
        """
        input:
            pred: prediction result from the model
        output:
            output: hand gesture like "one", "two", "three", "stone"...
        """
        return self.mapping[pred[0]]

    def normalize(self, df):
        if self.model.get_num_vertices() == 75: # hand + pose
            df = normalize_data(df)
        else:
            df = normalize_hand_data(df)
        return df.values


from collections import deque
from .control import MouseControl, KeyControl

class Module:

    def __init__(self, name, detector):
        self.name = name
        self.detector = detector
        self.mapping = {}
        self.actions = deque()
        self.dfs = deque()
        self.max_data_len = 60
        self.max_action_len = 1
        
    def add_mapping(self, control, actions):
        if isinstance(actions, str):
            actions = [actions]

        if len(actions) > self.max_action_len:
            self.max_action_len = len(actions)

        self.mapping[str(actions)] = control

    def add_transition(self, control, actions):
        self.add_mapping(control, actions)

    def add_mouse_mapping(self, control, action, **params):
        self.add_mapping(MouseControl(control, **params), action)

    def add_key_mapping(self, control, action):
        self.add_mapping(KeyControl(control), action)

    # actions to control
    def __call__(self, df):

        action = self.detector(df)
        self.update_actions(action)
        self.update_dfs(df)

        if len(self.actions) < self.max_action_len:
            return False

        for action_len in range(self.max_action_len, 0, -1):
            actions = [self.actions[i] for i in range(len(self.actions)-action_len,
                                 len(self.actions))]
            actions_str = str(actions)
            if actions_str in self.mapping:
                control = self.mapping[actions_str] # control can be a module to transit between modules
                
                if hasattr(control, "method_name") and "move_diff" in control.method_name:
                    if control.method_name == "right_move_diff":
                        fix_points = ["r_w"]
                    if control.method_name == "left_move_diff":
                        fix_points = ["l_w"]

                    fix_points_cols_x = [c + "_x" for c in fix_points]
                    fix_points_cols_y = [c + "_y" for c in fix_points]
                    control.set_params(df_data_1_x = self.dfs[-2][fix_points_cols_x], 
                                       df_data_2_x = self.dfs[-1][fix_points_cols_x],
                                       df_data_1_y = self.dfs[-2][fix_points_cols_y],
                                       df_data_2_y = self.dfs[-1][fix_points_cols_y])

                return control

        return False

    def update_actions(self, action):
        self.actions.append(action)
        if len(action) > self.max_data_len:
            self.actions.pop_left()

    def update_dfs(self, df):
        self.dfs.append(df)
        if len(df) > self.max_data_len:
            self.dfs.pop_left()

    def reset(self):
        self.actions = deque()
        self.dfs = deque()
        


from setuptools import setup, find_packages
import pkg_resources
import pathlib

with pathlib.Path('requirements.txt').open() as requirements_txt:
    requirements = [str(requirement) for requirement \
                    in pkg_resources.parse_requirements(requirements_txt)]

setup(
    name='pyhac',
    description='A human action controller running on different platforms',
    license='Apache License 2.0',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=requirements,
    include_package_data=True,
    author='JAQQ',
    author_email='dabit-lucas@gmail.com, chen.jiunhan@gmail.com',
    packages=find_packages(),
    keywords='human action controller',
    url='https://github.com/dabit-lucas/hac'
)



import mouse
import threading
import time
import keyboard
import sys
import pyautogui

if sys.platform.startswith("win"):
    import pydirectinput
    pydirectinput.PAUSE = 0
    pydirectinput.FAILSAFE = False
else:
    pyautogui.PAUSE = 0

class MouseControl:

    freeze = False
    def __init__(self, method, **params):
        self.execute = getattr(self, method)
        self.method_name = method
        self.sensitivity_factor = 4.0

        for key in params:
            setattr(self, key, params[key])

    def _check_freeze(func):
        def wrap(self, **params):
            sec = 0.01
            if not self.freeze:
                self.freeze = True
                func(self, **params)
                t = threading.Thread(target=self.freeze_timer, daemon=True, args=(sec,))
                t.start()

        return wrap

    @_check_freeze
    def click(self):
        if sys.platform.startswith("win"):
            mouse.click()
        else:
            pyautogui.click()
    @_check_freeze
    def mouse_left_down(self):
        if sys.platform.startswith("win"):
            mouse.press()
        else:
            pyautogui.mouseDown()
    @_check_freeze
    def mouse_left_up(self):
        if sys.platform.startswith("win"):
            mouse.release()
        else:
            pyautogui.mouseUp()
    @_check_freeze
    def mouse_right_down(self):
        if sys.platform.startswith("win"):
            mouse.press(button="right")
        else:
            pyautogui.mouseDown(button="right")
    @_check_freeze
    def mouse_right_up(self):
        if sys.platform.startswith("win"):
            mouse.release(button="right")
        else:
            pyautogui.mouseUp(button="right")
    @_check_freeze
    def right_click(self):
        if sys.platform.startswith("win"):
            mouse.right_click()
        else:
            mouse.right_click()
    @_check_freeze
    def roll_up(self):
        print("roll_up")
        if sys.platform.startswith("win"):
            mouse.wheel(30) 
        else:
            pyautogui.scroll(30) 
    @_check_freeze
    def roll_down(self):
        print("roll_down")
        if sys.platform.startswith("win"):
            mouse.wheel(-30)
        else:
            pyautogui.scroll(-30)
    @_check_freeze
    def move_diff(self):
        print("move_diff")
        fix_factor = 1000.0
        x1 = self.df_data_1_x.values.mean()
        x2 = self.df_data_2_x.values.mean()
        y1 = self.df_data_1_y.values.mean()
        y2 = self.df_data_2_y.values.mean()
        dx = -(x2 - x1).item() * self._sensitivity_factor * fix_factor
        dy = -(y2 - y1).item() * self._sensitivity_factor * fix_factor
        print(self._sensitivity_factor)
        if sys.platform.startswith("win"):
            pydirectinput.move(int(dx), int(dy))
        else:
            mouse.move(int(dx), int(dy), absolute=False)

    def right_move_diff(self):
        self.move_diff()

    def left_move_diff(self):
        self.move_diff()

    def set_params(self, **params):
        for key in params:
            setattr(self, key, params[key]) 

    def freeze_timer(self, sec):
        time.sleep(sec)
        self.freeze = False

    def release(self):
        self.mouse_left_up()
        self.mouse_right_up()

    @property
    def sensitivity_factor(self):
        return self._sensitivity_factor

    @sensitivity_factor.setter
    def sensitivity_factor(self, sensitivity_factor):
        self._sensitivity_factor = sensitivity_factor

class KeyControl:
    def __init__(self, key, **params):
        self.key = key
        self.method_name = key

    def execute(self):

        if sys.platform.startswith("win"):
            if self.key == "left":
                pydirectinput.keyDown('left')
            elif self.key == "right":
                pydirectinput.keyDown('right')
            else:
                keyboard.press(self.key)
        else:
            pyautogui.keyDown(self.key)
    def release(self):

        if sys.platform.startswith("win"):
            keyboard.release(self.key)
        else:
            pyautogui.keyUp(self.key)

